program mc48;
uses crt;

const version='v0.41a';  (* NUR ZU TESTZWECKEN! *)

(*
   Aenderungen: Binaerwerte-Darstellung 0011b oder %0011
                Hex-Darstellung 0Fh oder $0F
                Inline wie bisher immer Hex, ohne Kennung $ bzw. h

   v0.27: Fehler behoben: Hex-Zahlenwerte mit Endung "B" wurden als
          Binaerzahlen behandelt und ergaben Fehlermeldung.
          Z.B.: $4B wurde zur fehlerhaften Binaerzahl $4

          Neu: Anweisung "ORG #" erzeugt Pagewechsel auf die
          naechste Page.

   v0.29: Neuer Befehlscode FL (Forever-Loop) fuer
          zusaetzliche Befehle in MC48.TAB:
          Forever  HS      ;Wie bei While-Schleife wird die aktuelle
                           ;Adresse fuer den Ruecksprung auf den Anfang
                           ;dieser Schleife gesichert und der Schleifen-
                           ;Stack erhoeht.
          Again    JS FL   ;JS baut den Sprung zum Schleifenanfang ein.
                           ;Bei einer While-Schleife wuerde jetzt ein JE
                           ;den Sprung zum verlassen der Schleife setzen
                           ;und den Schleifenstack decrementieren.
                           ;In diesem Fall gibt es kein Verlassen der
                           ;Schleife, es muss lediglich der Stack korri-
                           ;giert werden, dafuer neuer Befehlscode FL.
          Zweck: Fuer die Endlos-Schleife des Hauptprogramms braucht keine
                 WhileNotDone-Schleife verwendet werden.

   v0.30: 2-Pass ! Prozeduren, welche im 1. Durchlauf nicht zugeordnet
          werden konnten, weil sie nach dem Aufruf definiert sind,
          werden im 2.PASS eingetragen.
          Damit ist es moeglich, Prozeduren in der Page-3 (also auch
          hinter dem Hauptprogramm) von vorherigen Pages aus aufzurufen.

          Aenderung: EProm-Puffer wird mit $00 initialisiert, damit der
          Disassembler bei ORG-Luecken keine MOV A,R7 eintraegt ...

   v0.31: Bereichsueberpruefung (0..255) nach GetVal-Aufrufen

   v0.32: Neu: Label- und GoTo-Anweisung
            Label bezeichner
            ...
            GoTo  bezeichner
          Neue Codes: LA, GO

   v0.33: Assembler-Anweisung ASM
   v0.34: ASM Sprungadressen ueberpruefen

   v0.35b:Kein generelles NOPpen des Page-Restes wenn die Seite $F0
          erreicht hat. Bei vielen Commands ist ein Seitenwechsel
          innerhalb der Codes zulaessig.
          In der TAB-Libraray sind fuer Page-Breaks kritische Codes
          mit einem '#' in der ersten Spalte gekennzeichnet. Bei diesen
          Codes wird nach dem Einsetzen geprueft, ob ein Pagewechsel
          aufgetreten ist. War dies der Fall, wird der BinCounter zurueck-
          gesetzt, der Rest der Seite mit NOPs gefuellt (PAGE-BREAK) und
          der Code auf der neuen Seite erneut erzeugt.
          Pagewechsel= wenn nach dem Einsetzen der BinCounter auf > $x01
          steht. Steht er auf $x00, so hat der Code bis auf $xFF der alten
          Seite gereicht.
          Neue TAB-Datei mit Markierung der kritischen Codes = MC48N.TAB.
          Erste Spalte, direkt vor Code '#'=kritisch, alles andere ausser
          Blank ist unkritischer Code. Blanks nicht zulaessig!
          Kritischer Code: enthaelt Page-abhaengige Sprungbefehle,
          z.B.: bedingte Spruenge jnz.., decjnz..
          Dies sind die If-Commands, While-Schleifen, Loops
   v0.35c:WriteBin-Prozedure: Abfrage ob EpromSize erreicht.
          Fehler in macLA: OpenLoop-Geschichte entfernt!
          Fehler in macAS-CheckPage: hi(bincounter+1) muss verglichen
          werden!
   v0.36: DATA-String-Anweisung: DATA "string",0
          ACHTUNG: Semikolon im String nicht zulaessig!
   v0.37: FEHLER: Vektoren fuer HardInt und TimerInt waren vertauscht!
   v0.38: Label werden mit ":LABEL" bezeichnet, nicht mehr mit ": LABEL"
          Kein LA-Eintrag mehr im Lib-File.
          Sprung zum Label mit "GOTO LABEL" (ohne ":")
   v0.39: Fehler im Assembler. MOV @Rx,A wurde Code 5x statt Ax   ###01
          Ursache: falsche Angabe im Elektor-Buch von Weidner.
   v0.40: Assembleranweisungen werden wie normale Befehle ( jetzt ohne
          Punkt) eingegeben. AS-Eintrag entfaellt.
          Problem unter Win95: _message.msg kann nicht geoeffnet werden.
          IOresult wird jetzt ignoriert.
          Erweiterung der Assemblercodes fuer 8021/22 mit Warnhinweis.
   v0.41: Die bei der VAR-Definition vergebenen Zuordnungsnummern sollten
          der Speicheradresse entsprechen, also mit 30 ($20) beginnen und
          nicht mit 0. Dies ist wichtig bei der Assemblerprogrammierung:
          MOV R1,#VARIABLE ist sonst nicht moeglich.
          Anpassungen in der TAB erforderlich!
          Zur Kompatibilitaet mit M48 muessen die Mem-Adressen (Zahlenwerte)
          angepasst werden (+32)!
   v0.41a:Bei binaerer Zahlendarstellung nur 4- und 8-stellige Werte zu-
          lassen, ansonsten Warnung ausgeben. (Fehlergefahr).
          Underlines in Binaerzahl zulaessig, z.B.: %11_001_101 %0011_1100
          Problem weiterhin: obige Warnung erscheint nicht in Listfile bei
          Konstanten-Definitionen!
*)


type str2=string[2]; str4=string[4]; str31=string[31];
     str63=string[80]; (******)

type cmdstr=^str31;  codstr=^str63;  procstr=^str31;  defstr=^str31;
const maxCommands=255; maxProcedures=64; maxReturns=64; maxWhiles=64;
      maxDefines=255;

    BinOffset=$09;
    EpromSize=4096;
    maxErrors=100;
    RAMoffset=$20; {erste verfuegbare RAM-Zelle (Mem)}

var commands:array[0..maxCommands] of cmdstr;
    codes   :array[0..maxCommands] of codstr;
    prozeduren:array[0..maxProcedures] of procstr;
    prozaddr:array[0..maxProcedures] of word;
    defines :array[0..maxDefines] of defstr; varcounter, defcounter:integer;
    defval  :array[0..maxDefines] of byte;
    ReturnStack:array[0..maxReturns] of word;
    WhileStack:array[0..maxWhiles] of word;
    JumpStack: array[0..maxWhiles] of word;
    OpenLoops: array[0..maxWhiles] of byte; OpenLoopCounter:byte;
    src, log, err, adr :text;
    LibCount:byte;
    ErrorCounter, SourceLine:integer;
    BinCounter, BinCounterMax, ResetVector:integer;
    ProcCounter, ReturnLevel, WhileLevel:byte;
    TableStart:byte; {zur Ueberwachung auf Page-Break}

    Eprom:array[0..EpromSize-1] of byte;

    {Ueberspringen des folgenden Befehls oder der folgenden Prozedur}
    {bei einer IF-Anweisung, Platzhalter NN}
    NNaktiv:boolean; NNadresse:word;
    inprozedur:boolean; {Prozedur wurde noch nicht mit EndProc abgeschl.}

    pass, undefined :byte;
    trycode:boolean; pagecheck: array[0..maxCommands] of boolean;

function upstr(s:str63):str63;
var i:integer;
begin
  if length(s)>0 then for i:=1 to length(s) do s[i]:=upcase(s[i]);
  upstr:=s;
end;

function hexbyte(b:byte):str2;
const hex:string[16]='0123456789ABCDEF';
begin
  hexbyte:=hex[(b shr 4)+1]+hex[(b and $0F)+1];
end;

function hexword(w:word):str4;
begin
  hexword:=hexbyte(hi(w))+hexbyte(lo(w));
end;

procedure FatalError(msg:string);
begin
  writeln('Fatal Error: ',msg);
  halt(1);
end;

procedure Fehler(msg:string);
begin
  inc(ErrorCounter);
  writeln('### FEHLER in Zeile',sourceline:4,': ',msg);
  write(log,#13,#10,'### FEHLER',ErrorCounter:3,' in Zeile',sourceline:4,'  ',msg);
  writeln(err,'Zeile:',sourceline:4,'  ',msg);
end;

procedure Warnung(msg:string);
begin
  writeln('### WARNUNG: ',msg);
  writeln(log,'### WARNUNG: ',msg);
  writeln(err,'! ',msg);
end;


function ReadTable:integer;
{ Einlesen der Code-Tabelle, liefert -1 bei Fehler oder Anzahl der Codes }
var tab:text;
    s:string; cmdstr {,codestr}:str63;
    TableCounter, i,j:integer;
begin
  TableCounter:=-1;
  assign(tab,'MC48n.TAB'); reset(tab);
  while not eof(tab) do begin
    readln(tab,s);
    if s<>'' then begin
      while (s[0]>#0) and (s[1]=' ') do delete(s,1,1);       {fuehrende Leerzeichen entf.}
      for i:=1 to length(s) do s[i]:=upcase(s[i]);  {nur Grossbuchstaben}
      while pos(#9,s)>0 do s[pos(#9,s)]:=' ';{TABs in einzelnes Leerzeichen}
      while (pos(';',s)>0) do dec(s[0]);     {Kommentar abschneiden}
      while s[length(s)]=' ' do dec(s[0]);   {angehaengte Leerz. entfernen}
      if (s<>'') then begin
        cmdstr:=copy(s,1,pos(' ',s)-1);
        delete(s,1,pos(' ',s));
        while (s[0]>#0) and (s[1]=' ') do delete(s,1,1);
        {
        codestr:='';
        for i:=1 to length(s) do begin
          codestr:=codestr+s[i];
        end;
        }
        inc(TableCounter);
        if (TableCounter<=MaxCommands) then begin
          if cmdstr[1]='#' then pagecheck[TableCounter]:=true
                           else pagecheck[TableCounter]:=false;
          delete(cmdstr,1,1);
          getmem(commands[TableCounter],sizeof(cmdstr));
          getmem(codes[TableCounter],sizeof(s)); {war codestr}
          commands[TableCounter]^:=cmdstr;
          codes[TableCounter]^:=s; {codestr;}
        end;
      end;
    end;
  end;
  close(tab);
  ReadTable:=TableCounter;
end;

procedure writebin(b:byte);
begin
  if Bincounter<EpromSize then begin
    EProm[BinCounter]:=b;
  {writeln(log,hexword(bincounter),':','$',hexbyte(b),' ');}
    write(log,' ',hexbyte(b));
    if binCounterMax<bincounter then binCounterMax:=binCounter;
  end else fehler('Maximale Codegroesse ueberschritten!');
  inc(binCounter);
end;

function getParameter(cmd:string):str63;
var s:str63;
begin
  s:='';
  if pos(' ',cmd)>0 then begin {Trennzeichen zum Parameter}
    while (cmd[0]>#0) and (cmd[1]<>' ') do delete(cmd,1,1);
    while (cmd[0]>#0) and (cmd[1]=' ') do delete(cmd,1,1);
    s:=cmd;
  end;
  getParameter:=s;
end;

function isDefined(s:str31):integer;
var i:integer;
begin
  i:=-1;
  repeat
    inc(i);
  until (i>DefCounter) or (upstr(defines[i]^)=upstr(s));
  if (i>DefCounter) then i:=-1;
  isDefined:=i;
end;

function isLib(s:str31):integer;
var i:integer;
begin
  i:=-1;
  repeat
    inc(i);
  until (i>LibCount) or (commands[i]^=upstr(s));
  if (i>LibCount) then i:=-1;
  isLib:=i;
end;

function isProc(s:str31):integer;
var i:integer;
begin
  i:=-1;
  repeat
    inc(i);
  until (i>ProcCounter) or (upstr(prozeduren[i]^)=upstr(s));
  if (i>ProcCounter) then i:=-1;
  isProc:=i;
end;

procedure getVal(s:str63; var wert,err:integer);
var i,m:integer;
begin
  {zuerst pruefen ob der String als VAR oder CONST definiert wurde}
  i:=isDefined(s);
  if (i>=0) then begin {es ist bereits Definiert}
    wert:=defval[i]; err:=0;
  end else begin
  if upcase(s[1]) in ['$','%','0'..'9','A'..'F',''''] then begin
    if (s[1]='''') and (s[3]='''') then begin
      wert:=ord(s[2]); err:=0;
    end else
    if (s[1]='$') or (upcase(s[length(s)])='H') then begin
      if upcase(s[length(s)])='H' then begin
        dec(s[0]); s:='$'+s;
      end;
      val(s,wert,err);
    end else
    if (s[1]='%') or (upcase(s[length(s)])='B') then begin
      if s[1]='%' then delete(s,1,1);
      if upcase(s[length(s)])='B' then delete(s,length(s),1); {dec(s[0]);}
      while pos('_',s)>0 do delete(s,pos('_',s),1);
      m:=1; wert:=0; err:=0;
      if (length(s)<>8) and (length(s)<>4) then
        Warnung('Binaerzahl sollte 8- od. 4-stellig eingegeben werden!');
      for i:= length(s) downto 1 do begin
        if (s[i]<>'1') and (s[i]<>'0') then err:=i
        else if s[i]='1' then wert:=wert+m;
        m:=m*2;
      end;
    end else
    begin
      val(s,wert,err);
    end;
  end;
  end;
end;

procedure GetAddr(s:str63; var w,e:integer);
var i:integer;
begin
  e:=255; w:=0;
  i:=isProc(s);
  if i>=0 then begin
    w:=ProzAddr[i]; e:=0;
  end else GetVal(s,w,e);
end;

procedure macPA(cmd:string);
var wert,err:integer; s:str63;
begin
  s:=getParameter(cmd);
  if s<>'' then begin
    getval(s,wert,err);
    if (err=0) and (wert>=0) and (wert<=255) then begin
      writebin($23); writebin(wert); {mov A,#XX}
    end else begin
      fehler('Ungueltiger Zahlenwert: ~'+s+'~');
    end;
  end;
end;

procedure macXX(cmd:string);
var wert,err:integer; s,s2:str63;
begin
  s:=getParameter(cmd); {Wert kann 1.Parameter...}
  s2:=getParameter(s);  {oder, wenn existiert, 2.Parameter sein}
  if s2<>'' then s:=s2;
  if s<>'' then begin
    getval(s,wert,err);
    if (err=0) and (wert>=0) and (wert<=255) then begin
      writebin(wert); {Konstante eintragen}
    end else begin
      fehler('Ungueltiger Zahlenwert: ~'+s+'~');
    end;
  end else fehler('Parameter fehlt');
end;

procedure macRR(cmd:string);
var wert,err:integer; s,s2:str63;
begin
  s:=getParameter(cmd); {Wert kann 1.Parameter...}
  s2:=getParameter(s);  {oder, wenn existiert, 2.Parameter sein}
  if s2<>'' then s:=s2;
  if s<>'' then begin
    getval(s,wert,err);
    if (err=0) and (wert>=$20) and (wert<=$FF) then begin
      writebin(wert); {MemSpeicher Konstante eintragen}
    end else begin
      fehler('unbekannte od. unzulaessige Variable ~'+s+'~');
    end;
  end else fehler('Parameter fehlt');
end;

procedure macPD(cmd:string);
var s:str31; i:integer;
begin
  if PASS=2 then exit;
  s:=getParameter(cmd);
  if s<>'' then begin
    {ueberpruefen ob Proz-Name als Befehl definiert ist}
    if isLib(s)>=0 then begin {es ist ein Befehl}
       fehler('Unzulaessiger Prozedurname ~'+s+'~');
    end else begin
      {ueberpruefen ob bereits als VAR oder CONST definiert}
      if isDefined(s)>=0 then begin {es ist bereits Definiert}
        fehler('~'+s+'~ wurde bereits als VAR/CONST definiert');
      end else begin
        {ueberpruefen ob bereits als Prozedur definiert}
        if isProc(s)>=0 then begin {es ist bereits Definiert}
          fehler('Prozedur/Label ~'+s+'~ existiert bereits');
        end else begin
          if InProzedur then
          if pass=1 then fehler('Procedure oder Interrupt nicht geschlossen!');
          InProzedur:=true;
          getmem(prozeduren[ProcCounter],sizeof(s));
          prozeduren[ProcCounter]^:=s;
          prozAddr[ProcCounter]:=BinCounter;
          if ProcCounter<MaxProcedures then inc(ProcCounter)
          else fehler('Zuviele Prozeduren/Label');

          write(log,' Prozedure definiert: [',ProcCounter,'] ~',s,'~ $',hexword(bincounter));
          {writeln(log,'+++ OpenLoopCounter=',OpenLoopCounter);}
          OpenLoops[OpenLoopCounter]:= WhileLevel;
          if OpenLoopCounter<MaxWhiles then inc(OpenLoopCounter)
          else fehler('Zuviele Schleifen');
        end;
      end;
    end;
  end else begin
    fehler('Fehler: Prozedurname fehlt');
  end;
end;

procedure defineLabel(cmd:string);
var s:str31; i:integer;
begin
  if PASS=2 then exit;
  s:=cmd;
  if s<>'' then begin
    {ueberpruefen ob Proz-Name als Befehl definiert ist}
    if isLib(s)>=0 then begin {es ist ein Befehl}
       fehler('Unzulaessiger Labelbezeichner ~'+s+'~');
    end else begin
      {ueberpruefen ob bereits als VAR oder CONST definiert}
      if isDefined(s)>=0 then begin {es ist bereits Definiert}
        fehler('~'+s+'~ wurde bereits als VAR/CONST definiert');
      end else begin
        {ueberpruefen ob bereits als Prozedur definiert}
        if isProc(s)>=0 then begin {es ist bereits Definiert}
          fehler('Prozedur/Label ~'+s+'~ existiert bereits');
        end else begin
          getmem(prozeduren[ProcCounter],sizeof(s));
          prozeduren[ProcCounter]^:=s;
          prozAddr[ProcCounter]:=BinCounter;
          if ProcCounter<MaxProcedures then inc(ProcCounter)
          else fehler('Zuviele Bezeichner/Prozeduren');
          write(log,' Label definiert: [',ProcCounter,'] ~',s,'~ $',hexword(bincounter));
        end;
      end;
    end;
  end else begin
    fehler('Fehler: Labelbezeichner fehlt');
  end;
end;



procedure macHA;
begin
  if InProzedur then
    fehler('Procedure oder Interrupt nicht geschlossen!');
  InProzedur:=true;

  write(log,' Jump zum Hauptprogramm eintragen: $',hexword(BinCounter));
  ResetVector:=BinCounter;

  Eprom[$0000]:=(hi(BinCounter) shl 5) or $04;
  Eprom[$0001]:=lo(BinCounter);

  {writeln(log,'+++ OpenLoopCounter=',OpenLoopCounter);}
  OpenLoops[OpenLoopCounter]:= WhileLevel;
  if OpenLoopCounter<MaxWhiles then inc(OpenLoopCounter)
  else fehler('Zuviele Schleifen');

end;

procedure macHH;
begin
  {zuerst feststellen, ob der Befehl zuvor ein CountX bzw. WrCountX war,
   dann ist die Registerbank bereits zurueckgesetzt (sel RB0, C5h) und
   dieser Befehl kann durch den gleichen dieser Routine ueberschrieben
   werden. Die Loop wird dann immer mit sel RB0 beginnen. ACHTUNG!
   auch ein Befehl mit Datenwert C5h (z.B. lade 250) wird als sel RB0
   gewertet. Da aber C5 mit C5 ueberschrieben wird, stoert dies nicht.
   Der Ruecksprung des Loop geht dann allerdings mitten in den Lade-
   befehl (auf das Datenbyte C5), die Funktion bleibt jedoch erhalten}
  {erstmal nicht aktiviert...
  if Eprom[BinCounter-1]=$C5 then dec(BinCounter);
  }

  ReturnStack[ReturnLevel]:=BinCounter;
  if ReturnLevel<MaxWhiles then inc(ReturnLevel)
  else if pass=1 then fehler('Zuviele Schleifen');

  {writeln(log,'### ReturnStack push $',hexword(bincounter));}
end;

procedure macSS;
var w:word;
begin
  {djnz-Ruecksprungadresse, 1Byte}
  if returnLevel>0 then dec(ReturnLevel)
  else if pass=1 then Fehler('EndWhile ohne WHILE');
  w:=ReturnStack[ReturnLevel];
  writebin(lo(w));

  {writeln(log,'### ReturnStack pop $',hexBYTE(LO(w)));}
end;

function macPR(cmd:string):str63;
var w,i:integer; s1,s2,s3:str63;
begin
  macPR:='';
  s1:=cmd;
  s2:=getParameter(s1);
  s3:=getParameter(s2);

  if s3<>'' then begin {kann nur Zahlenwert sein}
    s2:=copy(s2,1,pos(' ',s2)-1);
    while pos(' ',s2)>0 do delete(s2,pos(' ',s2),1);
    while pos(' ',s3)>0 do delete(s3,pos(' ',s3),1);
    i:=isLib(s2);
    if (isProc(s2)>=0) {or (pos('PA',codes[i]^)>0)} then begin
      getVal(s3,w,i);
      if (i=0) and (w>=0) and (w<=255) then begin
        macPA('* '+s3);
      end else begin
        fehler('~'+s3+'~ konnte nicht zugewiesen werden'); exit;
      end;
    end;
  end;

  if s2<>'' then begin {kann Parameter, Prozedur oder Befehl sein}
    s1:=copy(s1,1,pos(' ',s1)-1);
    while pos(' ',s1)>0 do delete(s1,pos(' ',s1),1);
    while pos(' ',s2)>0 do delete(s2,pos(' ',s2),1);
    i:=isLib(s2);
    if i>=0 then begin {es ist ein Befehl}
      s2:=codes[i]^;
      while pos(' ',s2)>0 do delete(s2,pos(' ',s2),1);
      macPR:=s2;
    end else
    if isLib(s1)>=0 then begin {s2 muss Prozedur sein}
      i:=isProc(s2);
      if i>=0 then begin {Prozedur aufrufen}
        w:=ProzAddr[i];
        writebin((hi(w) shl 5) or $14); {CALL}
        writebin(lo(w));
      end else begin
        inc(undefined);
        if PASS=1 then inc(Bincounter,2);
        if PASS=2 then begin
          fehler('Unbekannte Prozedur: ~'+s2+'~');
          exit;
        end;
      end;
    end else begin
      getVal(s2,w,i);
      if (i=0) and (w>=0) and (w<=255) then begin
        macPA('* '+s2);
      end else begin
        fehler('~'+s2+'~ konnte nicht zugewiesen werden'); exit;
      end;
    end;
  end;

  {s1 kann nur Befehl oder Prozedur sein}
  while pos(' ',s1)>0 do delete(s1,pos(' ',s1),1);
  i:=isProc(s1);
  if i>=0 then begin {Prozedur aufrufen}
    w:=ProzAddr[i];
    writebin((hi(w) shl 5) or $14); {CALL}
    writebin(lo(w));
  end else begin
    if isLib(s1)<0 then begin
      inc(undefined);
      if PASS=1 then inc(Bincounter,2);
      if PASS=2 then fehler('Befehl oder Prozedur unbekannt: ~'+s1+'~');
    end;
  end;
end;


procedure macGO(cmd:string);
var s1,s2:str63; i,w:integer;
begin
  s1:=GetParameter(cmd);   (* goto ziel *)
  s2:=GetParameter(s1);    (* ifA=0 goto ziel *)
  if s2<>'' then s1:=s2;
  if s1<>'' then begin
    i:=isProc(s1);
    if i>=0 then begin  (*ist ein Label*)
      w:=ProzAddr[i];
      writebin((hi(w) shl 5) or $04); {JMP}
      writebin(lo(w));
    end else begin      (*Label nicht gefunden, Adresse?*)
      GetVal(s1,w,i);
      if i=0 then begin (*gueltige Adressangabe*)
        writebin((hi(w) shl 5) or $04); {JMP}
        writebin(lo(w));
      end else begin
        inc(undefined);
        if PASS=1 then inc(Bincounter,2); (*Platzhalter*)
        if PASS=2 then fehler('Ungueltiges Sprungziel ~'+s1+'~');
      end;
    end;
  end else fehler('Sprungziel fehlt: ~'+cmd+'~');
end;

procedure macJP;
begin
  (****)
  writebin((hi(BinCounter) shl 5) or $04);
end;

procedure macHS;
begin
  WhileStack[WhileLevel]:=BinCounter;
  if WhileLevel<MaxWhiles then inc(WhileLevel)
  else fehler('Zuviele WHILE-Schleifen');
  {writeln(log,'### WhileStack push $',hexword(bincounter));}
end;

procedure macJS;
var w:word;
begin
  w:=WhileStack[WhileLevel-1];
  writebin((hi(w) shl 5) or $04);
  writebin(lo(w));

  {writeln(log,'### WhileStack pop $',hexword(w));}
end;

procedure macJ3;
begin
  writebin(lo(BinCounter+3));
end;

procedure macJZ;
begin
  JumpStack[WhileLevel]:=BinCounter-1;
  {writeln(log,'### WhileAdr. gespeichert:',hexword(JumpStack[whilelevel]));}
end;

procedure macJE;
begin
  EProm[JumpStack[WhileLevel]]:=lo(BinCounter);
  {JP-Jump korrigieren, da ev. Page-Wechsel : }
  EProm[JumpStack[WhileLevel]-1]:=(hi(BinCounter) shl 5) or $04;
  {writeln(log,'### WhileAddr:',hexbyte(lo(BinCounter)),' Eprom:',hexword(JumpStack[WhileLevel]));}
  if WhileLevel>0 then dec(WhileLevel) else fehler('EndWhile ohne WHILE');
end;

procedure macFL; {###v0.29 Forever-Loop}
begin
  if WhileLevel>0 then dec(WhileLevel) else fehler('EndWhile ohne WHILE');
end;

procedure macSO;
{### Pruefen ob Schleife offen}
begin
  if OpenLoopCounter>0 then dec(OpenLoopCounter);
  if (OpenLoops[OpenLoopCounter]<>WhileLevel) then
{fehler('### Loop-Stack ERROR ! '+hexbyte(OpenLoopCounter)+'='+hexbyte(OpenLoops[OpenLoopCounter])+'<>'+hexbyte(WhileLevel))}
    fehler('Offene Schleife')
  else writeln(log);{writeln(log,'### Loop-Stack OK ')};
  if not(InProzedur) then
    if PASS=1 then fehler('Prozedur- oder Interrupt-Block nicht geoeffnet!');
  InProzedur:=false;
end;


procedure macHX(cmd:string);
var s:str63; ws:str2; wert,err:integer;
begin
  s:=GetParameter(cmd);
  if s<>'' then begin
    while s[1]=' ' do delete(s,1,1);
    while length(s)>1 do begin
      ws:=copy(s,1,2); delete(s,1,2);
      if s[1]=',' then delete(s,1,1);
      val('$'+ws,wert,err);
      if err=0 then begin
        writebin(wert);
      end else begin
        fehler('Unzulaessiger HEX-Wert: ~'+ws+'~');
      end;
    end;
  end else begin
    fehler('Keine Daten bei Inline-Anweisung');
  end;
end;

procedure macDX(cmd:string);
var s:str63; ws:str63; wert,err,i:integer;
begin
  s:=GetParameter(cmd);
  if s<>'' then begin
    while length(s)>0 do begin
      while (s[0]>#0) and ((s[1]=' ') or (s[1]=',')) do delete(s,1,1);
      if (pos('"',s)=1) then begin  {"string"}
        delete(s,1,1);
        if (pos('"',s)>0) then begin {string-ende suchen}
          ws:=copy(s,1,pos('"',s)-1);
          delete(s,1,pos('"',s));
          for i:=1 to length(ws) do begin
            getval(''''+ws[i]+'''',wert,err); {kein Fehler moeglich..}
            writebin(wert);
          end;
        end else fehler('String nicht abgeschlossen: ~"'+s+'~');
      end else begin
        if (pos(',',s)>1) then begin
          ws:=copy(s,1,pos(',',s)-1); delete(s,1,pos(',',s));
        end else begin
          ws:=s; s:='';
        end;
        {while (pos(' ',ws)=1) do delete(ws,1,1);}                {###v036}
        while ws[length(ws)]=' ' do delete(ws,length(ws),1);    {###v036}
        if (length(ws)>0) then begin
          getval(ws,wert,err);
          if (err=0) and (wert>=0) and (wert<=255) then begin
            writebin(wert);
          end else begin
            fehler('Unzulaessiger Wert: ~'+ws+'~');
          end;
        end;
      end;
    end;
  end else begin
    fehler('Keine Daten bei Data-Anweisung');
  end;
end;


procedure macTI;
begin
  if InProzedur then
   fehler('Procedure oder Interrupt nicht geschlossen!');
  InProzedur:=true;

  write(log,' Timer-Interruptvektor eintragen: $',hexword(BinCounter));
  EProm[$07]:=(hi(BinCounter) shl 5) or $04;
  EProm[$08]:=lo(BinCounter);
    {writeln(log,'+++ OpenLoopCounter=',OpenLoopCounter);}
  OpenLoops[OpenLoopCounter]:= WhileLevel;
  if OpenLoopCounter<MaxWhiles then inc(OpenLoopCounter)
  else fehler('Zuviele Schleifen');

end;

procedure macHI;
begin
  if InProzedur then
    fehler('Procedure oder Interrupt nicht geschlossen!');
  InProzedur:=true;
  write(log,' External-Interruptvektor eintragen: $',hexword(BinCounter));
  EProm[$03]:=(hi(BinCounter) shl 5) or $04;
  EProm[$04]:=lo(BinCounter);
  {writeln(log,'+++ OpenLoopCounter=',OpenLoopCounter);}
  OpenLoops[OpenLoopCounter]:= WhileLevel;
  if OpenLoopCounter<MaxWhiles then inc(OpenLoopCounter)
  else fehler('Zuviele Schleifen');
end;

procedure macNN;
begin
  NNaktiv:=true; NNadresse:=BinCounter;
  writeBin($00); {Platzhalter fuer Byte}
end;

procedure macXO(cmd:string); {ORG-Anweisung}
var s:str31; wert,err:integer;
begin
  s:=getParameter(cmd); err:=0;
  if (s='#') then begin
    wert:=bincounter;
    if lo(bincounter)>0 then wert:=(wert+256) and $FF00;
  end else getVal(s,wert,err);
  if (err<>0) or (wert<0) then begin
    fehler('Unkorrekte ORG-Anweisung  ~'+cmd+'~');
  end else begin
    if (wert>EpromSize-1) then begin
      fehler('Ausserhalb des Adressbereiches ~'+cmd+'~');
    end else begin
      if (wert<bincounter) then
        warnung('Adressbereich ist bereits belegt ~'+cmd+'~');
        bincounter:=wert;
    end;
  end;
end;

function macAS(cmd:string): boolean;
var s1,s2:str63; b,b0,b1,b2:integer;  i,err:integer;

  procedure asmfehler;
  begin
    fehler('Unkorrekter Assemblercode ~'+cmd+'~');
  end;

  procedure adrfehler;
  begin
    if PASS=1 then inc(undefined)
              else fehler('Ungueltiges Sprungziel ~'+cmd+'~');
  end;

  procedure checkpage(jadr:integer);
  begin
    if hi(bincounter+1)<>hi(jadr) then fehler('Sprungadresse ausserhalb der Page');
  end;

  procedure checkjump(jadr:integer);
  begin
    if jadr>$7FF then warnung('Sprungadresse ausserhalb 2k-Bereich');
  end;

begin
  macAS:=true;
  fillchar(s1,sizeof(s1),#0); fillchar(s2,sizeof(s2),#0);
  s1:=upstr(GetParameter('ASM '+cmd));
  s2:=GetParameter(s1);
  if s2<>'' then s1:=copy(s1,1,pos(' ',s1)-1);
  if s1=''  then fehler('Assemblerbefehl fehlt');
  b0:=1; {1-Byte-Befehl, Vorgabe} b1:=$00; b2:=$00;

  if s1='ADD' then begin
    if s2='' then asmfehler else
    if pos('A,R',s2)=1 then begin
      b:=ord(s2[4])-$30;
      if b in [0..7] then b1:=$68 or b else asmfehler;
    end else
    if pos('A,@R',s2)=1 then begin
      b:=ord(s2[5])-$30;
      if b in [0,1] then b1:=$60 or b else asmfehler;
    end else
    if pos('A,#',s2)=1 then begin
      b0:=2; b1:=$03;
      delete(s2,1,3); GetVal(s2,b2,err); if err>0 then asmfehler;
    end else asmfehler;
  end else

  if s1='ADDC' then begin
    if s2='' then asmfehler else
    if pos('A,R',s2)=1 then begin
      b:=ord(s2[4])-$30;
      if b in [0..7] then b1:=$78 or b else asmfehler;
    end else
    if pos('A,@R',s2)=1 then begin
      b:=ord(s2[5])-$30;
      if b in [0,1] then b1:=$70 or b else asmfehler;
    end else
    if pos('A,#',s2)=1 then begin
      b0:=2; b1:=$13;
      delete(s2,1,3); GetVal(s2,b2,err); if err>0 then asmfehler;
    end else asmfehler;
  end else

  if s1='ANL' then begin
    if s2='' then asmfehler else
    if pos('A,R',s2)=1 then begin
      b:=ord(s2[4])-$30;
      if b in [0..7] then b1:=$58 or b else asmfehler;
    end else
    if pos('A,@R',s2)=1 then begin
      b:=ord(s2[5])-$30;
      if b in [0,1] then b1:=$50 or b else asmfehler;
    end else
    if pos('A,#',s2)=1 then begin
      b0:=2; b1:=$53;
      delete(s2,1,3); GetVal(s2,b2,err); if err>0 then asmfehler;
    end else
    if pos('BUS,#',s2)=1 then begin
      b0:=2; b1:=$98;
      delete(s2,1,5); GetVal(s2,b2,err); if err>0 then asmfehler;
    end else
    if pos('P1,#',s2)=1 then begin
      b0:=2; b1:=$99;
      delete(s2,1,4); GetVal(s2,b2,err); if err>0 then asmfehler;
    end else
    if pos('P2,#',s2)=1 then begin
      b0:=2; b1:=$9A;
      delete(s2,1,4); GetVal(s2,b2,err); if err>0 then asmfehler;
    end else asmfehler;
  end else

  if s1='ANLD' then begin
    if s2='P4,A' then b1:=$9C else
    if s2='P5,A' then b1:=$9D else
    if s2='P6,A' then b1:=$9E else
    if s2='P7,A' then b1:=$9F else asmfehler;
  end else

  if s1='CALL' then begin
    b0:=2;
    GetAddr(s2,i,err);
    if err>0 then adrfehler else checkjump(i);
    b1:= (hi(i) shl 5) or $14;
    b2:=lo(i);
  end else

  if s1='CLR'  then begin
    if s2='A'  then b1:=$27 else
    if s2='C'  then b1:=$97 else
    if s2='F1' then b1:=$A5 else
    if s2='F0' then b1:=$85 else asmfehler;
  end else

  if s1='CPL' then begin
    if s2='A'  then b1:=$37 else
    if s2='C'  then b1:=$A7 else
    if s2='F1' then b1:=$B5 else
    if s2='F0' then b1:=$95 else asmfehler;
  end else

  if s1='DA' then begin
    if s2='A' then b1:=$57 else asmfehler;
  end else

  if s1='DEC' then begin
    if s2='A' then b1:=$07 else
    if pos('R',s2)=1 then begin
      b:=ord(s2[2])-$30;
      if b in [0..7] then b1:=$C8 or b else asmfehler;
    end else asmfehler;
  end else

  if s1='DIS' then begin
    if s2='I' then b1:=$15 else
    if s2='TCNTI' then b1:=$35 else asmfehler;
  end else

  if s1='DJNZ' then begin
    b0:=2;
    if pos('R',s2)=1 then begin
      b:=ord(s2[2])-$30; b1:=$E8 or b;
      delete(s2,1,3); GetAddr(s2,b2,err);
      if err>0 then adrfehler else checkpage(b2);
    end else asmfehler;
  end else

  if s1='EN' then begin
    if s2='I' then b1:=$05 else
    if s2='TCNTI' then b1:=$25 else asmfehler;
  end else

  if s1='ENT0' then begin
    if s2='CLK' then b1:=$75 else asmfehler;
  end else

  if s1='IN' then begin
    if s2='A,P1' then b1:=$09 else
    if s2='A,P2' then b1:=$0A else
    if s2='A,P0' then begin
      b1:=$08; warnung('ASM-Befehl nur fuer 8021,8022');
    end else asmfehler;
  end else

  if s1='INC' then begin
    if s2='A' then b1:=$17 else
    if pos('R',s2)=1 then begin
      b:=ord(s2[2])-$30;
      if b in [0..7] then b1:=$18 or b else asmfehler;
    end else
    if pos('@R',s2)=1 then begin
      b:=ord(s2[3])-$30;
      if b in [0..7] then b1:=$10 or b else asmfehler;
    end;
  end else

  if s1='INS' then begin
    if s2='A,BUS' then b1:=$08 else asmfehler;
  end else

  if pos('JB',s1)=1 then begin
    b0:=2;
    b:=ord(s1[3])-$30;
    if b in [0..7] then b1:=(b shl 5) or $12 else asmfehler;
    GetAddr(s2,b2,err); if err>0 then adrfehler else checkpage(b2);
  end else

  if s1='JC' then begin
    b0:=2; b1:=$F6;
    GetAddr(s2,b2,err); if err>0 then adrfehler else checkpage(b2);
  end else

  if s1='JF0' then begin
    b0:=2; b1:=$B6;
    GetAddr(s2,b2,err); if err>0 then adrfehler else checkpage(b2);
  end else

  if s1='JF1' then begin
    b0:=2; b1:=$76;
    GetAddr(s2,b2,err); if err>0 then adrfehler else checkpage(b2);
  end else

  if s1='JMP' then begin
    b0:=2;
    GetAddr(s2,i,err); if err>0 then adrfehler else checkjump(i);
    b1:= (hi(i) shl 5) or $04;
    b2:=lo(i);
  end else

  if s1='JMPP' then begin
    if s2='@A' then b1:=$B3 else asmfehler;
  end else

  if s1='JNC' then begin
    b0:=2; b1:=$E6;
    GetAddr(s2,b2,err); if err>0 then adrfehler else checkpage(b2);
  end else

  if s1='JNI' then begin
    b0:=2; b1:=$86;
    GetAddr(s2,b2,err); if err>0 then adrfehler else checkpage(b2);
  end else

  if s1='JNT0' then begin
    b0:=2; b1:=$26;
    GetAddr(s2,b2,err); if err>0 then adrfehler else checkpage(b2);
  end else

  if s1='JNT1' then begin
    b0:=2; b1:=$46;
    GetAddr(s2,b2,err); if err>0 then adrfehler else checkpage(b2);
  end else

  if s1='JNZ' then begin
    b0:=2; b1:=$96;
    GetAddr(s2,b2,err); if err>0 then adrfehler else checkpage(b2);
  end else

  if s1='JTF' then begin
    b0:=2; b1:=$16;
    GetAddr(s2,b2,err); if err>0 then adrfehler else checkpage(b2);
  end else

  if s1='JT0' then begin
    b0:=2; b1:=$36;
    GetAddr(s2,b2,err); if err>0 then adrfehler else checkpage(b2);
  end else

  if s1='JT1' then begin
    b0:=2; b1:=$56;
    GetAddr(s2,b2,err); if err>0 then adrfehler else checkpage(b2);
  end else

  if s1='JZ' then begin
    b0:=2; b1:=$C6;
    GetAddr(s2,b2,err); if err>0 then adrfehler else checkpage(b2);
  end else

  if s1='MOV' then begin
    if pos('A,#',s2)=1 then begin
      b0:=2; b1:=$23;
      delete(s2,1,3); GetVal(s2,b2,err); if err>0 then asmfehler;
    end else
    if s2='A,PSW' then begin
      b1:=$C7;
    end else
    if pos('A,R',s2)=1 then begin
      b:=ord(s2[4])-$30;
      if b in [0..7] then b1:=$F8 or b else asmfehler;
    end else
    if s2='A,@R0' then b1:=$F0 else
    if s2='A,@R1' then b1:=$F1 else
    if s2='A,T'   then b1:=$42 else
    if s2='PSW,A' then b1:=$D7 else
    if pos('R',s2)=1 then begin
      if s2[4]='A' then begin
        b:=ord(s2[2])-$30;
        if b in [0..7] then b1:=$A8 or b else asmfehler;
      end else
      if s2[4]='#' then begin
        b0:=2;
        b:=ord(s2[2])-$30;
        if b in [0..7] then b1:=$B8 or b else asmfehler;
        delete(s2,1,4); GetVal(s2,b2,err); if err>0 then asmfehler;
      end else asmfehler;
    end else
    if s2='@R0,A' then b1:=$A0 else    (* ###01 *)
    if s2='@R1,A' then b1:=$A1 else    (* ###01 *)
    if pos('@R0,#',s2)=1 then begin
      b0:=2; b1:=$B0;
      delete(s2,1,5); GetVal(s2,b2,err); if err>0 then asmfehler;
    end else
    if pos('@R1,#',s2)=1 then begin
      b0:=2; b1:=$B1;
      delete(s2,1,5); GetVal(s2,b2,err); if err>0 then asmfehler;
    end else
    if s2='T,A' then b1:=$62 else asmfehler;
  end else

  if s1='MOVD' then begin
    if s2='A,P4' then b1:=$0C else
    if s2='A,P5' then b1:=$0D else
    if s2='A,P6' then b1:=$0E else
    if s2='A,P7' then b1:=$0F else
    if s2='P4,A' then b1:=$3C else
    if s2='P5,A' then b1:=$3D else
    if s2='P6,A' then b1:=$3E else
    if s2='P7,A' then b1:=$3F else asmfehler;
  end else

  if s1='MOVP' then begin
    if s2='A,@A' then b1:=$A3 else asmfehler;
  end else

  if s1='MOVP3' then begin
    if s2='A,@A' then b1:=$E3 else asmfehler;
  end else

  if s1='MOVX' then begin
    if s2='A,@R0' then b1:=$80 else
    if s2='A,@R1' then b1:=$81 else
    if s2='@R0,A' then b1:=$90 else
    if s2='@R1,A' then b1:=$01 else asmfehler;
  end else

  if s1='NOP' then b1:=$00 else

  if s1='ORL' then begin
    if s2='' then asmfehler else
    if pos('A,R',s2)=1 then begin
      b:=ord(s2[4])-$30;
      if b in [0..7] then b1:=$48 or b else asmfehler;
    end else
    if pos('A,@R',s2)=1 then begin
      b:=ord(s2[5])-$30;
      if b in [0,1] then b1:=$40 or b else asmfehler;
    end else
    if pos('A,#',s2)=1 then begin
      b0:=2; b1:=$43;
      delete(s2,1,3); GetVal(s2,b2,err); if err>0 then asmfehler;
    end else
    if pos('BUS,#',s2)=1 then begin
      b0:=2; b1:=$88;
      delete(s2,1,5); GetVal(s2,b2,err); if err>0 then asmfehler;
    end else
    if pos('P1,#',s2)=1 then begin
      b0:=2; b1:=$89;
      delete(s2,1,4); GetVal(s2,b2,err); if err>0 then asmfehler;
    end else
    if pos('P2,#',s2)=1 then begin
      b0:=2; b1:=$8A;
      delete(s2,1,4); GetVal(s2,b2,err); if err>0 then asmfehler;
    end else asmfehler;
  end else

  if s1='ORLD' then begin
    if s2='P4,A' then b1:=$8C else
    if s2='P5,A' then b1:=$8D else
    if s2='P6,A' then b1:=$8E else
    if s2='P7,A' then b1:=$8F else asmfehler;
  end else

  if s1='OUTL' then begin
    if s2='BUS,A' then b1:=$02 else
    if s2='P1,A'  then b1:=$39 else
    if s2='P2,A'  then b1:=$3A else
    if s2='P0,A'  then begin
      b1:=$90; warnung('ASM-Befehl nur fuer 8021,8022');
    end else asmfehler;
  end else

  if s1='RAD' then begin
    b1:=$80; warnung('ASM-Befehl nur fuer 8022');
  end else

  if s1='RET' then b1:=$83 else

  if s1='RETR' then b1:=$93 else

  if s1='RETI' then begin
    b1:=$93; warnung('ASM-Befehl nur fuer 8022');
  end else

  if s1='RL' then begin
    if s2='A' then b1:=$E7 else asmfehler;
  end else

  if s1='RLC' then begin
    if s2='A' then b1:=$F7 else asmfehler;
  end else

  if s1='RR' then begin
    if s2='A' then b1:=$77 else asmfehler;
  end else

  if s1='RRC' then begin
    if s2='A' then b1:=$67 else asmfehler;
  end else

  if s1='SEL' then begin
    if s2='MB0' then b1:=$E5 else
    if s2='MB1' then b1:=$F5 else
    if s2='RB0' then b1:=$C5 else
    if s2='RB1' then b1:=$D5 else
    if s2='AN0' then begin
      b1:=$95; warnung('ASM-Befehl nur fuer 8022');
    end else
    if s2='AN1' then begin
      b1:=$85; warnung('ASM-Befehl nur fuer 8022');
    end else asmfehler;
  end else

  if s1='STOP' then begin
    if s2='TCNT' then b1:=$65 else asmfehler;
  end else

  if s1='STRT' then begin
    if s2='T'   then b1:=$55 else
    if s2='CNT' then b1:=$45 else asmfehler;
  end else

  if s1='SWAP' then begin
    if s2='A' then b1:=$47 else asmfehler;
  end else

  if s1='XCH' then begin
    if pos('A,R',s2)=1 then begin
      b:=ord(s2[4])-$30;
      if b in [0..7] then b1:=$28 or b else asmfehler;
    end else
    if s2='A,@R0' then b1:=$20 else
    if s2='A,@R1' then b1:=$21 else asmfehler;
  end else

  if s1='XCHD' then begin
    if s2='A,@R0' then b1:=$30 else
    if s2='A,@R1' then b1:=$31 else asmfehler;
  end else

  if s1='XRL' then begin
    if s2='' then asmfehler else
    if pos('A,R',s2)=1 then begin
      b:=ord(s2[4])-$30;
      if b in [0..7] then b1:=$D8 or b else asmfehler;
    end else
    if pos('A,@R',s2)=1 then begin
      b:=ord(s2[5])-$30;
      if b in [0,1] then b1:=$D0 or b else asmfehler;
    end else
    if pos('A,#',s2)=1 then begin
      b0:=2; b1:=$D3;
      delete(s2,1,3); GetVal(s2,b2,err); if err>0 then asmfehler;
    end else asmfehler;
  end else macAS:=false;
  write(log,'  ',hexword(bincounter),':');
  writebin(lo(b1)); if b0=2 then writebin(lo(b2));
end;

procedure DefineVar(s:string);
var s2:str31; idx:string[5]; wert,err:integer;

  procedure RegisterVar;
  begin
    if (isDefined(s2+idx)<0) and
       (isLib(s2+idx)<0)     and
       (isProc(s2+idx)<0) then begin
      getmem(defines[defcounter],sizeof(s2)+sizeof(idx));
      defines[defcounter]^:=s2+idx;
      defval[defcounter]:=varcounter+RAMoffset;
      if defcounter<maxDefines then begin
        inc(defcounter); inc(varcounter);
      end else fehler('Zuviele Definitionen (CONST, VAR)');
    end else fehler('Bezeichner doppelt oder reserviert: ~'+s2+idx+'~');
  end;

begin
  if PASS=2 then exit;
  {VAR XCOUNT, TEST, XYZ}
  if (pos(' ',s)>0) then delete(s,1,pos(' ',s)) else s:='';
  while length(s)>0 do begin
    while (s[0]>#0) and ((s[1]=' ') or (s[1]=',')) do delete(s,1,1);
    if pos(',',s)>0 then s2:=copy(s,1,pos(',',s)-1)
                    else s2:=s;
    delete(s,1,length(s2));
    while s2[length(s2)]=' ' do dec(s2[0]);
    if (pos(']',s2)>0) and (pos('[',s2)>0) and ((pos(']',s2))>(pos('[',s2))) then
    begin
      val(copy(s2,pos('[',s2)+1,pos(']',s2)-pos('[',s2)-1),wert,err);
      if (err=0) and (wert>0) and (wert<256) {###} then begin
        s2:=copy(s2,1,pos('[',s2)-1);
        for err:=0 to wert do begin
          str(err,idx);  idx:='['+idx+']';
          RegisterVar;
        end;
      end else fehler('Array-Definition ~'+s2+'~');
    end else begin
      idx:='';
      RegisterVar;
    end;
  end;
end;

procedure DefineConst(s:string);
var s2,s3:str31; wert,err:integer;
begin
  if PASS=2 then exit;
  {DEF XCOUNT=12, TEST=2, XYZ=$33}
  if (pos(' ',s)>0) then delete(s,1,pos(' ',s)) else s:='';
  while length(s)>0 do begin
    while (s[0]>#0) and ((s[1]=' ') or (s[1]=',')) do delete(s,1,1);
    if pos(',',s)>0 then s2:=copy(s,1,pos(',',s)-1)
                    else s2:=s;
    delete(s,1,length(s2));
    while s2[length(s2)]=' ' do dec(s2[0]);
    s3:=copy(s2,1,pos('=',s2)-1);
    delete(s2,1,length(s3)+1);
    while s3[length(s3)]=' ' do dec(s3[0]);
    while s2[1]=' ' do delete(s2,1,1);
    getval(s2,wert,err);
    if (isDefined(s3)<0) and
       (isLib(s3)<0)     and
       (isProc(s3)<0) then begin
      if (err=0) and (wert>=0) and (wert<256) then begin
        getmem(defines[defcounter],sizeof(s3));
        defines[defcounter]^:=s3;
        defval[defcounter]:=wert;
        if defcounter<maxDefines then begin
          inc(defcounter);
        end else fehler('Zuviele Definitionen (CONST, VAR)');

      end else fehler('Wertzuweisung ~'+s3+'~=~'+s2+'~');
    end else fehler('Bezeichner doppelt oder reserviert: ~'+s3+'~');
  end;
end;

procedure DefineTable(s:string);
begin
  if (pos(' ',s)>0) then delete(s,1,pos(' ',s)) else s:='';
  s:='DEF '+s+'=$'+hexbyte(lo(bincounter));
  DefineConst(s); {CONST tablename=$xx}
  TableStart:=lo(BinCounter);
end;

procedure AnalyzeCode(cmd:string;i:integer);
var s:str63; mac:string[2]; cstr:str63;
    wert,err:integer;
begin

  cstr:=cmd;  s:=codes[i]^;
  while pos(' ',s)>0 do delete(s,pos(' ',s),1);
  while (length(s)>=2) do begin
    mac:=copy(s,1,2); delete(s,1,2);
    if mac ='PA' then begin macPA(cstr);{getKonstante} end else
    if mac ='XX' then begin macXX(cstr); end else
    if mac ='RR' then begin macRR(cstr); end else
    if mac ='PD' then begin macPD(cstr); {ProzedurDefinition} end else
    (* if mac ='LA' then begin macLA(cstr); {Label definieren} end else *)
    if mac ='HA' then begin macHA; {Hauptprogramm-Anfang} end else
    if mac ='HH' then begin macHH; {AktuelleAdresseMerken} end else
    if mac ='SS' then begin macSS; {RuecksprungEintragen} end else
    if mac ='PR' then begin s:=macPR(cstr)+s; {Befehl/Prozeduraufruf} end else
    if mac ='NN' then begin macNN; {**** SprungadrBerechnen} end else
    if mac ='JP' then begin macJP; {JMP_Page} end else
    if mac ='HS' then begin macHS; {IncWhile} end else
    if mac ='JS' then begin macJS; {WhileRuecksprung} end else
    if mac ='J3' then begin macJ3; {Jump-3-weiter} end else
    if mac ='JZ' then begin macJZ; {TemporaereSprungAdr} end else
    if mac ='JE' then begin macJE; {TempSprungAdrEinsetzen} end else
    if mac ='FL' then begin macFL; {###v0.29} end else
    if mac ='SO' then begin macSO; {OffeneSchleifenPruefen} end else
    if mac ='HX' then begin macHX(cstr); {InlineCodeEinsetzen} end else
    if mac ='DX' then begin macDX(cstr); {Data-Anweisung} end else
    if mac ='TI' then begin macTI; {TimerInterruptProz} end else
    if mac ='HI' then begin macHI; {HardwareIntProz} end else
    if mac ='XO' then begin macXO(cstr); {*NEU ORG-Anweisung} end else
    if mac ='VA' then begin DefineVAR(cstr); end else
    if mac ='CO' then begin DefineConst(cstr); end else
    if mac ='TA' then begin DefineTable(cstr); end else
    if mac ='GO' then begin macGO(cstr);{GoTo} end else
    (*if mac ='AS' then begin macAS(cstr);{Assembler} end else*)
    begin {normaler HEX-Code}
      val('$'+mac,wert,err);
      if err=0 then begin
        writebin(wert);
      end else begin
        fehler('Unbekannter Code: ~'+mac+'~ in ~'+codes[i]^+'~');
      end;
    end;
  end;

 {Bei einer IF-Anweisung wird hier die Adresse zum Ueberspringen}
 {des folgenden Befehls/Prozedur eingtragen}
 if NNaktiv then begin
   EProm[NNadresse]:=lo(bincounter);
   NNaktiv:=false;
 end;

end;

procedure PageBreak;
begin
   writeln(log,#13,#10,'  ',hexword(BinCounter),': === PAGE-BREAK ===');
   writeln(adr,hexword(BinCounter),' === PAGE-BREAK ===');
   write(log,'  ',hexword(bincounter),':');
   while(lo(BinCounter)<>$00) do writeBin($00); {NOPs einsetzen}
end;

procedure Translate(s:string);
label tryagain;
var i:integer; cmd:str63; codeover:boolean;
    tryaddr,tryWlevel, tryRlevel:integer;
begin
  if pos(' ',s)>0 then cmd:=copy(s,1,pos(' ',s)-1)
                  else cmd:=s;

  if (cmd[1]=':') then begin {es ist ein Label}
    writeln(log,#13,#10,'source:~',s,'~');
    write(log,'  ',hexword(bincounter),':');
    delete(cmd,1,1);
    defineLabel(cmd);
    exit;
  end;

  i:=isLib(cmd); {ist es ein Befehl?}

  codeover:=false; trycode:=pagecheck[i];
  tryaddr:=bincounter; tryWlevel:=whilelevel; tryRlevel:=returnlevel;

tryagain:

  if codeover then PageBreak;

  if not codeover then writeln(log,#13,#10,'source:~',s,'~')
                  else writeln(log,#13,#10,' moved:~',s,'~');

  if (i>=0) then begin  {es ist ein Befehl (.TAB) }
    writeln(log,'  code: ',codes[i]^);
    write(log,'  ',hexword(bincounter),':');
    AnalyzeCode(s,i);
  end else begin        {ist es ein Prozeduraufruf?}
    if (isProc(cmd)>=0) then begin
      write(log,'  ',hexword(bincounter),':');
      if macPR(s)<>'' then begin
        fehler('Fehler???  ~'+s+'~');
      end;
    end else begin      {ist es eine ASM-Anweisung? }
      if not macAS(s) then begin  {nein: unbekannter Befehl}
        inc(undefined);
        if PASS=1 then inc(bincounter,2); (* Platzhalter fer PASS-2 *)
        if PASS=2 then fehler('Befehl oder Prozedur unbekannt: ~'+cmd+'~');
      end;
    end;
  end;
  (***)
  {Pruefen ob Table Page ueberschreitet}
  if (i>=0) and (pos('DX',codes[i]^)>0) then
    if (TableStart>lo(bincounter)) then begin
      fehler('Page-Break innerhalb Data-Anweisung!');
      TableStart:=lo(bincounter);{fehler zuruecksetzen}
    end;

  if trycode
    and (hi(bincounter)>hi(tryaddr)) and (lo(bincounter)>$00)
    then begin
    codeOver:=true;
    BinCounter:=TryAddr;
    WhileLevel:=tryWlevel;
    ReturnLevel:=tryRlevel;
    trycode:=false;
    goto tryagain;
  end;

end;

function OpenFiles(s:string):byte;
{ Quelltext-Datei oeffnen }
begin
{$i-}
  assign(src,s+'.M48'); reset(src);
  if ioresult=0 then
    assign(log,s+'.LST'); rewrite(log);
  if ioresult=0 then
  assign(adr,s+'.ADR'); rewrite(adr);
  OpenFiles:=ioresult;
  if ioresult=0 then
  assign(err,'MC48N.ERR'); rewrite(err);
{$i+}
end;

function CloseFiles:byte;
begin
{$i-}
  close(src); close(log); close(err); close(adr);
{$i+}
  CloseFiles:=ioresult;
  {if ErrorCounter=0 then erase(err);}
end;

procedure WriteEpromFile(s:string);
var f:file;
begin
  assign(f,s+'.BIN'); rewrite(f,1);
  {BlockWrite(f,EProm,SizeOf(EProm));}
  BlockWrite(f,EProm,BinCounterMax+1);
  Close(f);
end;


procedure do_it;
var i:byte; s:string; incfile:text; IncludeFile:boolean;

  procedure OpenInclude;
  begin
{$i-}
    assign(incfile,getParameter(s)); reset(incfile);
{$i+}
    if (ioresult=0) then begin
      IncludeFile:=true;
      write(log,#13,#10,'<<< OPEN INCLUDE-FILE ',getParameter(s),#13,#10);
    end
    else fehler('Include-File nicht gefunden: '+getParameter(s));
  end;

begin
  writeln(#13#10,'### PASS: ',pass,#13#10);
  BinCounter:=BinOffset; BinCounterMax:=BinCounter; ResetVector:=0;
  ErrorCounter:=0; OpenLoopCounter:=0;
  ReturnLevel:=0; WhileLevel:=0; SourceLine:=0;
  IncludeFile:=false; InProzedur:=false;
  if pass=1 then begin
    varCounter:=0; defCounter:=0; ProcCounter:=0;
  end;
  fillchar(eprom,sizeof(eprom),$00);
  while (not eof(src)) and (ErrorCounter<MaxErrors) do begin
    if IncludeFile then begin
      if eof(incfile) then begin
        close(incfile); IncludeFile:=false;
        writeln(log,#13,#10,'>>> CLOSE INCLUDE-FILE');
        s:='';
      end else begin
        readln(incfile,s);
      end;
    end else begin
      readln(src,s); inc(sourceLine);
      write(sourceline:3,#8#8#8);
    end;
    {writeln('### FILE-INPUT ~',s,'~');}
    writeln(adr,hexword(bincounter),' ',s);
    while (s[0]>#0) and (pos(#9,s)>0) do s[pos(#9,s)]:=' '; {TABs in ein Blank wandeln}
    while (S[0]>#0) and (pos(';',s)>0) do dec(s[0]); { Kommentar abschneiden }
    while (s[0]>#0) and (s[1]=' ') do delete(s,1,1); { fuehrende Blanks entfernen }
    while (s[0]>#0) and (s[length(s)]=' ') do dec(s[0]); { angehaengte Blanks entfernen }
    if (length(s)>0) then begin
      if pos('INCLUDE ',upstr(s))=1 then OpenInclude else
      begin
        NNaktiv:=false;
        Translate(s);
      end;
    end;
  end;
  writeln(log,#13,#10,#13,#10);
end;

var dummy:byte;
begin
  clrscr;
  writeln('### MC48N ',version,'/hi');
  writeln('### Datei: ',paramstr(1));
  LibCount:=ReadTable;
  if LibCount>MaxCommands then FatalError('TAB-Datei: zuviele Eintraege!')
                          else writeln('### Befehlssatz: ',LibCount,' Eintraege');
  if LibCount=0 then FatalError('TAB-Datei enthaelt keine Eintraege');
  dummy:=OpenFiles(paramstr(1));
  if dummy<>0 then FatalError('Dateifehler ~'+paramstr(1)+'~');
  writeln(log,'### MC48N ',version,'/hi');

  PASS:=1; undefined:=0;
  do_it;
  PASS:=2;
  if (undefined>0) and (ErrorCounter=0) then begin
    GotoXY(1,whereY);
    writeln('### ',undefined,' fehlende Zuordnung/en. PASS-2 wird gestartet...');
    dummy:=CloseFiles; dummy:=OpenFiles(paramstr(1));
    do_it;
  end;

  if varCounter>96 then fehler('RAM fuer 8049 ueberschritten!')
  else if varCounter>32 then warnung('RAM fuer 8048 ueberschritten!');

  if bincounter>$0FFF then fehler('Programmgroesse ueber 4k!') else
  if bincounter>$07FF then warnung('Programmgroesse ueber 2k!');

  if ErrorCounter=0 then WriteEpromFile(paramstr(1));

  if (ProcCounter>0) then begin
    writeln(log,'*** Prozedur- und Label-Adressen:');
    for dummy:=0 to ProcCounter-1 do
      writeln(log,'   $',hexword(ProzAddr[dummy]),' ~',Prozeduren[dummy]^,'~');
  end;

  writeln(log,'   $',hexword(ResetVector),' ~Begin~ (Reset-Vector)');

  if (DefCounter>0) then begin
    writeln(log,'*** Defines:');
    for dummy:=0 to DefCounter-1 do
      writeln(log,'   ~',defines[dummy]^,'~ = ($',hexbyte(defval[dummy]),') ',defval[dummy]);
  end;
  if varCounter>0 then begin
    write(log,'*** RAM-Belegung: ',varCounter);
    if (VarCounter>1) then write(log,' Bytes') else write(log,' Byte');
    writeln(log,' (RAM-Adresse $',hexbyte(VarCounter+31),')');
  end;
  if (errorCounter=0) then begin
    writeln(err,'* Compilation OK !  Keine Fehler');
    writeln(err,'* Codesize:  ',BinCounter:4,' Bytes (ROM-Location $',hexword(bincounter-1),')');
    writeln(err,'* Variables: ',varCounter:4,' Bytes (RAM-Location $',hexbyte(VarCounter+31),')');
    writeln('### Codesize: ',BinCounter,' Bytes');
    writeln(log,'*** Codesize: ',BinCounter,' Bytes (ROM-Location $',hexword(bincounter-1),')');
    sound(4000); delay(40); NoSound;
  end else begin
    writeln('### Keine Code-Datei erzeugt');
    sound(400); delay(250); NoSound;
  end;
  writeln('### Fehler: ',ErrorCounter);
  writeln(log,#13,#10,#13,#10,'*** Fehler: ',ErrorCounter);

  dummy:=CloseFiles;

  if ErrorCounter>0 then halt(255);
end.

