;
; This S/R shifts the register pair whose LSB
;   is being pointed at by R1 right one position
; All registers are preserved
PROCEDURE ROTATERIGHTR1:
  CLR C
:SHIFTRIGHTR1:
  INC R1          ;R1 points to the LSB
  MOV     A,@R1   ;get MSB
  RRC     A
  MOV     @R1,A   ;Result MSB
  DEC     R1      ;point at LSB
  MOV     A,@R1   ;get it
  RRC     A
  MOV     @R1,A   ;Store result LSB
EndProc
;
;-----------------------------------------------
;
; This S/R shifts the register pair whose LSB is
;  being indicated by R0 left one position
; All registers are preserved
PROCEDURE ROTATELEFTR0:
  CLR C
:SHIFTLEFTR0:
  MOV A,@R0       ;get LSB
  RLC     A
  MOV     @R0,A   ;Result LSB
  INC     R0      ;Point at MSB
  MOV     A,@R0
  RLC     A
  MOV     @R0,A   ;store MSB
  DEC     R0      ;restore R0
EndProc
;
;---------------------------------------------------
;
; This S/R, ADD, adds a 16 Bit number pointed at by R0
;  to a sum indicated by R1The sum (@R1) is updated
;  @R0 is not changed
; All registers preserved
PROCEDURE ADD:
  CLR     C       ;Add without carry, or with it:
:ADDC:
  MOV     A,@R1   ;LSB of sum
  ADDC    A,@R0   ; + LSB of number
  MOV     @R1,A   ;New sum LSB to @R1
  INC     R1
  INC     R0      ;point at MSBs
  MOV     A,@R1   ;MSB of sum
  ADDC    A,@R0   ; + MSB of number
  MOV     @R1,A   ;New sum MSB to @R1
  DEC     R1      ;Restore R1, R0
  DEC     R0
EndProc
;
;----------------------------------------------------
;
;MULTIPLY16:  16 x 16 Bit multiplication
;
;result (@R1) = multiplicand (@R0) x multiplier (@R1)
;
; It overflows safely, but without indication
; Registers R4,R5,R6,R3 are destroyed
;
;       Example: 5 x 3
;               101
;            x  011
;       ----------------
;               101     (Shifted LSBit=1)
;              101      (Shifted LSBit=1)
;             000       (Shifted LSBit=0,no add)
;       ----------------
;             01111     (Result)
;
PROCEDURE MULTIPLY16
  MOV     R3,#16  ;Loop count in R3
  CLR     A       ;Clear result
  MOV     R6,A    ;R6 is result MSB
  MOV     R5,A    ;R5 is LSB
; shift and test if multiplier LSBit = 1
:ZZMULT:
  CALL    ROTATERIGHTR1: ;0-> MSBit,LSBit-> C
  JNC     ZZNOADD: ;LSBit not 1, Don't add
;
  MOV     A,R1    ;Save R1 in
  MOV     R4,A    ; R4
  MOV     R1,#5   ;Point at the result R5,R6
  CALL    ADD:    ;Add result, multiplicand @R0
  MOV     A,R4    ;Restore
  MOV     R1,A    ;R1
; (Could test for overflow here:)
;
;Now we shift left the multiplicand
:ZZNOADD: CALL ROTATELEFTR0: ;0->LSBit, MSBit->Carry
;
  DJNZ    R3,ZZMULT:      ;Loop for 16 Bits
; Now shift the result to @R1
  MOV     A,R0    ;Save R0
  MOV     R4,A    ;in R4
  MOV     R0,#5   ;Point to R5,R6
  MOV     A,@R0   ;Get LSB
  MOV     @R1,A   ;Put LSB
  INC     R1
  INC     R0
  MOV     A,@R0   ;Get MSB
  MOV     @R1,A   ;Put MSB
  DEC     R1      ;Restore R1
  MOV     A,R4    ;Restore R0
  MOV     R0,A
EndProc


